package net.sf.buildbox.util;

import java.io.IOException;

/**
 * Helper class for producers that are periodically scanning some input
 */
public abstract class AbstractSampler<T> {

    /**
     * Check if there is an object available and pull it.
     *
     * @return pulled object or null if none available.
     * @throws java.io.IOException implementation can use it to indicate IO problems
     * @throws InterruptedException -
     */
    protected abstract T scan() throws IOException, InterruptedException;

    /**
     * Wait some time for an object
     *
     * @param timeout total time limit; 0 means try just once, less than 0 means no time limit
     * @param quantum scanning period, should be much less than timeout
     * @return pulled object or null if none was available all the time.
     * @throws java.io.IOException propagated here from {@link #scan()} to indicate IO problems
     * @throws InterruptedException -
     */
    public T produceObject(long timeout, long quantum) throws IOException, InterruptedException {
        long startTime = System.currentTimeMillis();
        long time0 = startTime;
        long endTime = time0 + timeout;
        T object = scan();
        long scanTime = System.currentTimeMillis() - time0;
        if (timeout == 0) return object; // try once
        while (object == null) {
            long sleeptime;
            if (timeout < 0) {
                // infinite waiting - we don't check endTime here
                sleeptime = quantum;
            } else {
                sleeptime = endTime - System.currentTimeMillis();
                if (sleeptime > quantum) {
                    sleeptime = quantum;
                } else if (sleeptime <= 0) {
                    break;
                }
            }
            if (sleeptime > scanTime) {
                Thread.sleep(sleeptime - scanTime);
            }
            time0 = System.currentTimeMillis();
            object = scan();
            if (Thread.interrupted()) {
                final long time = System.currentTimeMillis() - startTime;
                throw new InterruptedException("interrupted after " + time + " millis");
            }
            scanTime = System.currentTimeMillis() - time0;
        }
        return object;
    }
}
