package net.sf.buildbox.util;

import java.text.ParseException;
import java.util.Map;
import java.util.Collection;
import java.util.HashMap;

public class BbxStringUtils {

    BbxStringUtils() {
        // not instantiable
    }

    // Constants specifying String align type
    public static final int ALIGN_LEFT = 0;
    public static final int ALIGN_RIGHT = 1;
    public static final int ALIGN_MIDDLE = 2;
    private static final String[] fAsciiNames = new String[]{
            "NUL", null, "STX", "ETX", // 00-03
            null, null, null, null, // 04-07
            null, "TAB", "LF", null, // 08-0b
            "FF", "CR", null, null,  // 0c-0f
            null, null, null, null, // 10-13
            null, null, null, null, // 14-17
            null, null, null, "ESC", // 18-1b
            null, null, null, null, // 1c-1f
    };

    /**
     * Align string to the left, right or middle.
     *
     * @param str  input string for aligning
     * @param mode constant specifying align mode (use ALIGN_LEFT, ALIGN_RIGHT or ALIGN_MIDDLE)
     * @param chr  character which to use for aligning
     * @param size minimal size of the resulting string
     * @return aligned string
     */
    public static String strAlign(String str, int mode, char chr, int size) {

        return strAlign(str, mode, chr, chr, size);
    }

    /**
     * Align string to the left, right or middle.
     *
     * @param str  input string for aligning
     * @param mode constant specifying align mode (use ALIGN_LEFT, ALIGN_RIGHT or ALIGN_MIDDLE)
     * @param chrL character which to use for aligning
     * @param chrR character which to use for aligning
     * @param size minimal size of the resulting string
     * @return aligned string
     */
    public static String strAlign(String str, int mode, char chrL, char chrR, int size) {

        if (str == null || str.length() >= size) {
            return str;
        }

        StringBuilder sb = new StringBuilder();
        int count = size - str.length();
        if (mode == ALIGN_LEFT) {
            sb.append(str);
            for (int i = 0; i < count; i++) {
                sb.append(chrR);
            }

        } else if (mode == ALIGN_RIGHT) {
            for (int i = 0; i < count; i++) {
                sb.append(chrL);
            }
            sb.append(str);

        } else if (mode == ALIGN_MIDDLE) {
            int countL = count / 2;
            int countR = count - countL;
            for (int i = 0; i < countL; i++) {
                sb.append(chrL);
            }
            sb.append(str);
            for (int i = 0; i < countR; i++) {
                sb.append(chrR);
            }

        } else {
            return str;
        }

        return sb.toString();
    }

    /**
     * Escapes string for using in HTML.
     *
     * @param s <code>String</code> to escape
     * @return escaped <code>String</code>
     */
    public static String escapeString(Object s) {

        String s1;
        if (s == null) {
            return null;
        }
        s1 = s.toString();
        StringBuilder sb = new StringBuilder(s1.length());
        for (int i = 0; i < s1.length(); i++) {
            char c;
            switch (c = s1.charAt(i)) {
            case '<':
                sb.append("&lt;");
                break;
            case '>':
                sb.append("&gt;");
                break;
            case '&':
                sb.append("&amp;");
                break;
            default:
                sb.append(c);
                break;
            }
        }
        return sb.toString();
    }

    /**
     * Decodes sequence of hexadecimal letters into sequence of bytes.
     * No delimiters are allowed
     *
     * @param s byte content to decode
     * @return array of bytes
     * @throws java.text.ParseException if parameter doesn't represent <code>byte</code> array
     */
    public static byte[] hexDecodeBytes(String s) throws ParseException {
        if (s == null) {
            return null;
        }
        if (s.length() % 2 != 0) {
            throw new ParseException("length is odd", s.length());
        }
        byte[] bytes = new byte[s.length() / 2];
        for (int i = 0; i < bytes.length; i++) {
            try {
                bytes[i] = (byte) Integer.parseInt(s.substring(2 * i, 2 * i + 2), 16);
            } catch (NumberFormatException e) {
                throw new ParseException(e.getMessage(), i * 2);
            }
        }
        return bytes;
    }

    /**
     * Encodes byte array to sequence of hexadecimal letters.
     * No delimiters are placed between subsequent bytes.
     *
     * @param bytes byte content to encode
     * @return encoded byte content
     */
    public static String hexEncodeBytes(byte[] bytes) {
        if (bytes == null) {
            return null;
        }
        StringBuilder sb = new StringBuilder();
        for (byte b : bytes) {
            String s = Integer.toHexString(b & 0xff);
            if (s.length() < 2) {
                sb.append("0");
            }
            sb.append(s);
        }
        return sb.toString();
    }

    public static String bytesToString(byte[] aBytes) {
        final StringBuilder sb = new StringBuilder();
        for (byte b : aBytes) {
            if ((b & 128) != 0) {
                sb.append("<0x");
                Integer.toString(b, 2);
                sb.append(Integer.toHexString(b + 256));
                sb.append('>');
            } else if (b == '<') {
                sb.append("<<>");
            } else if (b == '>') {
                sb.append("<>>");
            } else if (b >= 32) {
                sb.append((char) b);
            } else if (fAsciiNames[b] == null) {
                sb.append("<0x");
                sb.append(Integer.toHexString(b));
                sb.append('>');
            } else {
                sb.append('<');
                sb.append(fAsciiNames[b]);
                sb.append('>');
            }
        }
        return sb.toString();
    }


    public static interface ReferenceResolver {
        public String resolve(String reference);
    }

    private static Map<String,String> props = new HashMap<String, String>();

    static {
        for (Map.Entry<Object, Object> entry : System.getProperties().entrySet()) {
            props.put((String)entry.getKey(), (String) entry.getValue());
        }
        for (Map.Entry<String, String> entry : System.getenv().entrySet()) {
            props.put("env." + entry.getKey(), entry.getValue());
        }
    }

    //TODO: improve - lookup on demand, method name etc.
    public static String expandSysProps(String fileRef) {
        return BbxStringUtils.expandProps(fileRef, props);
    }

    /**
     * Expands template with properties, referenced using ANT syntax.
     * Properties not found in <code>props</code> parameter remain unexpanded.<br/>
     *
     * @param template the string to expand, for instance: <code>${JAVA_HOME}/lib/tools.jar</code>
     * @param props    properties to use; you can use {@link System#getenv()} here.
     * @return a string with all existing properties replaced with their values.
     */
    public static String expandProps(String template, Map<String, String> props) {
        return replaceTokens(template, props, "${", "}", null);
    }

    public static String replaceTokens(String template, Map<String, String> tokens, final String tokenDelimStart, final String tokenDelimEnd, final Collection<String> badNames) {
        class MapResolver implements ReferenceResolver {
            private final Map<String, String> map;

            public MapResolver(Map<String, String> map) {
                this.map = map;
            }

            public String resolve(String reference) {
                String result = map.get(reference);
                if (result == null) {
                    if (badNames != null) {
                        badNames.add(reference);
                    }
                    result = tokenDelimStart + reference + tokenDelimEnd;
                }
                return result;
            }
        }
        return resolveReferences(template, new MapResolver(tokens), tokenDelimStart, tokenDelimEnd);
    }

    public static String resolveReferences(String template, ReferenceResolver referenceResolver, String tokenDelimStart, String tokenDelimEnd) {
        final StringBuilder sb = new StringBuilder();
        int srcpos = 0;
        int n1 = template.indexOf(tokenDelimStart);
        while (n1 >= 0) {
            final int n2 = template.indexOf(tokenDelimEnd, n1);
            if (n2 < 0) break;
            sb.append(template.substring(srcpos, n1));
            final String name = template.substring(n1 + tokenDelimStart.length(), n2);
            final String value = referenceResolver.resolve(name);
            sb.append(value);
            srcpos = n2 + tokenDelimEnd.length();
            n1 = template.indexOf(tokenDelimStart,srcpos);
        }
        sb.append(template.substring(srcpos));
        return sb.toString();
    }

}
