package net.sf.buildbox.util;

import java.util.*;

public final class Iterables {
    private static ElementFilter<Object> NOT_NULL_FILTER = new ElementFilter<Object>() {
        public boolean accept(Object element) {
            return element != null;
        }
    };

    private Iterables() {
        // not instantiable
    }

    public static <T> Collection<T> collect(Iterable<T> coll) {
        final Collection<T> result = new LinkedList<T>();
        addAllTo(result, coll);
        return Collections.unmodifiableCollection(result);
    }

    public static <T> void addAllTo(Collection<T> result, Iterable<T> from) {
        for (T item : from) {
            result.add(item);
        }
    }

    /**
     * Returned iterator produces unique elements (no duplicates).
     * @param source  the source iterator that may contain duplicates
     * @return dedupped iterator
     */
    public static <T> Iterable<T> distinct(final Iterable<T> source) {
        final ElementFilter<T> elementFilter = new ElementFilter<T>() {
            private final HashSet<T> set = new HashSet<T>();
            public boolean accept(T element) {
                return set.add(element);
            }
        };
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return where(source.iterator(), elementFilter);
            }
        };
    }
    
    public static <T> Iterable<T> concat(Iterable<? extends Iterable<? extends T>> iterables) {
        final Iterator<? extends Iterable<? extends T>> iter = iterables.iterator();
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return new Concatenator<T>() {
                    protected Iterator<? extends T> nextIterator() {
                        Iterable<? extends T> iterable = iter.next();
                        return iterable == null ? null : iterable.iterator();
                    }

                    protected boolean hasNextIterator() {
                        return iter.hasNext();
                    }
                };
            }
        };
    }

    public static <T> Iterable<T> concat(final Iterable<? extends T>... iterables) {
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return new Concatenator<T>() {
                    private int nextIndex = 0;

                    protected boolean hasNextIterator() {
                        return nextIndex < iterables.length;
                    }

                    protected Iterator<? extends T> nextIterator() {
                        Iterable<? extends T> iterable = iterables[nextIndex++];
                        return iterable != null ? iterable.iterator() : null;
                    }
                };
            }
        };
    }

    private abstract static class Concatenator<T> implements Iterator<T> {
        Iterator<? extends T> currentIterator = null;

        protected abstract Iterator<? extends T> nextIterator();

        protected abstract boolean hasNextIterator();

        private void findReadyIterator() {
            while (true) {
                if (currentIterator != null) {
                    if (currentIterator.hasNext()) {
                        break;
                    }
                }
                if (!hasNextIterator()) {
                    break;
                }
                currentIterator = nextIterator();
            }
        }

        public boolean hasNext() {
            findReadyIterator();
            return currentIterator != null && currentIterator.hasNext();
        }

        public T next() {
            findReadyIterator();
            if (currentIterator != null && currentIterator.hasNext()) {
                return currentIterator.next();
            } else {
                throw new NoSuchElementException();
            }
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

    public static <T> Iterable<T> where(final Iterable<T> source, final ElementFilter<T> elementFilter) {
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return where(source.iterator(), elementFilter);
            }
        };
    }

    @SuppressWarnings("unchecked")
    public static <T> Iterable<T> whereNotNull(final Iterable<T> source) {
        return where(source, (ElementFilter<T>) NOT_NULL_FILTER);
    }

    private static <T> Iterator<T> where(final Iterator<T> sourceIterator, final ElementFilter<T> elementFilter) {
        return new Iterator<T>() {
            private T currentElement;
            private boolean ready = false;

            private boolean findReadyElement() {
                while (!ready) {
                    if (!sourceIterator.hasNext()) break;
                    currentElement = sourceIterator.next();
                    ready = elementFilter.accept(currentElement);
                }
                return ready;
            }

            public boolean hasNext() {
                return findReadyElement();
            }

            public T next() {
                if (!findReadyElement()) {
                    throw new NoSuchElementException(elementFilter.toString());
                }
                ready = false;
                return currentElement;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    public static interface ElementFilter<T> {
        public boolean accept(T element);
    }

    public abstract static class Deref<K,V> implements Iterable<V> {
      private Iterable<? extends K> keys;

      public Deref(Iterable<? extends K> keys) {
        if (keys == null) {
          throw new NullPointerException();
        }
        this.keys = keys;
      }

      public abstract V get(K key);

      public Iterator<V> iterator() {
        return new Iterator<V>() {
          private Iterator<? extends K> ki = keys.iterator();
          public boolean hasNext() {
            return ki.hasNext();
          }

          public V next() {
            return get(ki.next());
          }

          public void remove() {
            throw new UnsupportedOperationException();
          }
        };
      }
    }

}
