package net.sf.buildbox.util;

import java.io.File;
import java.io.FileFilter;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * Traverses a filesystem subtree and invokes specified callback on each directory and file element.
 * For callback, the {@link FileFilter} class is abused  - so the implementor has to return a boolean value (ignored)
 * Supports sorting, subtree exclussion, fileonly/dironly mode.
 * <p>The motivation for this approach is to have alternative to {@link org.codehaus.plexus.util.FileUtils#getFiles(java.io.File, String, String, boolean)} which is too
 * memory-intensive on big trees; also, it takes too long before producing first results, which might be bad for incremental processing</p>
 * @author Petr Kozelka <pkozelka@gmail.com>
 */
public class PushingDirectoryTraversal {
    private final FileFilter filter;
    private boolean sorted = false;
    private boolean wantFiles = true;
    private boolean wantDirs = false;
    private final Set<String> dirNameExcludes = new HashSet<String>();

    public PushingDirectoryTraversal(FileFilter filter) {
        this.filter = filter;
    }


    public void setSorted(boolean sorted) {
        this.sorted = sorted;
    }

    public void setWantFiles(boolean wantFiles) {
        this.wantFiles = wantFiles;
    }

    public void setWantDirs(boolean wantDirs) {
        this.wantDirs = wantDirs;
    }

    public void addDirectoryExclude(String dirNameExclude) {
        dirNameExcludes.add(dirNameExclude);
    }

    public void traverse(File subdir) {
        final File[] files = subdir.getAbsoluteFile().listFiles(new FileFilter() {
            public boolean accept(File pathname) {
                return pathname.isFile() ? wantFiles : canIncludeDir(pathname.getName());
            }
        });
        if (files == null) {
            System.err.println("ERROR: Directory does not exist: " + subdir);
            return;
        }
        if (sorted) {
            Arrays.sort(files);
        }
        for (File file : files) {
            if (file.isDirectory()) {
                if (wantDirs) {
                    filter.accept(file);
                }
                traverse(file);
            } else {
                filter.accept(file);
            }
        }
    }

    private boolean canIncludeDir(String name) {
        return ! dirNameExcludes.contains(name);
    }
}
