package net.sf.buildbox.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Gradually create several digests for a stream at once.
 * <p>
 * Usage:
 * <pre>
 * final MessageDigest md5 = MessageDigest.getInstance("MD5");
 * final MessageDigest sha1 = MessageDigest.getInstance("SHA1");
 * StreamingMultiDigester.compute(path, md5, sha1);
 * final String md5sum = BbxStringUtils.hexEncodeBytes(md5.digest());
 * final String sha1sum = BbxStringUtils.hexEncodeBytes(sha1.digest());
 * </pre>
 * </p>
 * @author <a href="mailto:pkozelka@gmail.com">Petr Kozelka</a>
 * @todo instead, implement multidigesting stream decorator
 */
public class StreamingMultiDigester {

    private final MessageDigest[] digests;

    private static final int BUFFER_SIZE = 32768;

    private StreamingMultiDigester(MessageDigest... digests) throws NoSuchAlgorithmException {
        this.digests = digests;
    }

    private void update(InputStream is) throws IOException {
        final byte[] buffer = new byte[BUFFER_SIZE];
        int size = is.read(buffer, 0, BUFFER_SIZE);
        while (size >= 0) {
            for (MessageDigest digest : digests) {
                digest.update(buffer, 0, size);
            }
            size = is.read(buffer, 0, BUFFER_SIZE);
        }
    }

    public static void compute(InputStream is, MessageDigest... digests) throws NoSuchAlgorithmException, IOException {
        final StreamingMultiDigester smd = new StreamingMultiDigester(digests);
        smd.update(is);
    }

    public static void compute(File file, MessageDigest... digests) throws NoSuchAlgorithmException, IOException {
        final InputStream is = new FileInputStream(file);
        try {
            final StreamingMultiDigester smd = new StreamingMultiDigester(digests);
            smd.update(is);
        } finally {
            is.close();
        }
    }
}
