package net.sf.buildbox.util.exec;

/*
 * Copyright The Codehaus Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


import org.codehaus.plexus.util.Os;
import org.codehaus.plexus.util.cli.*;

import java.io.InputStream;
import java.io.PrintWriter;
import java.util.concurrent.Callable;

/**
 * Cloned {@link CommandLineUtils#executeCommandLine(org.codehaus.plexus.util.cli.Commandline, org.codehaus.plexus.util.cli.StreamConsumer, org.codehaus.plexus.util.cli.StreamConsumer)}
 * allowing to customize the waiting operation; for instance, by waiting on a synchronized queue.
 * This allows creating managed threads simply.
 *
 * @author <a href="mailto:pkozelka@gmail.com">Petr Kozelka</a>
 */
public class CooperativeExec implements Callable<Integer> {
    public static final StreamConsumer DEV_STDOUT = new DefaultConsumer();
    public static final StreamConsumer DEV_STDERR = new WriterStreamConsumer(new PrintWriter(System.err));
    public static final StreamConsumer DEV_NULL = new StreamConsumer() {
        public void consumeLine(String line) {
        }
    };
    private final Commandline commandLine;
    private long timeout = -1;
    private long quantum = 100;
    private InputStream systemIn;
    private StreamConsumer systemOut = DEV_STDOUT;
    private StreamConsumer systemErr = DEV_STDERR;

    public CooperativeExec(Commandline commandLine) {
        this.commandLine = commandLine;
    }

    /**
     * On non-windows systems, changes shell to {@link StraightDummyShell} which adds no shell stuff arround.
     * You can suppress this behavior by overriding with empty method body.
     */
    protected void fixCommandLine() {
        if (!Os.isFamily( Os.FAMILY_WINDOWS )) {
            commandLine.setShell(new StraightDummyShell());
        }
    }

    public Integer call() throws CommandLineException {
        if (commandLine == null) {
            throw new IllegalArgumentException("commandLine cannot be null.");
        }
        fixCommandLine();
        final Process p = commandLine.execute();
        final StreamFeeder inputFeeder = systemIn != null ? new StreamFeeder(systemIn, p.getOutputStream()) : null;
        final StreamPumper outputPumper = new StreamPumper(p.getInputStream(), systemOut);
        final StreamPumper errorPumper = new StreamPumper(p.getErrorStream(), systemErr);

        if (inputFeeder != null) {
            inputFeeder.start();
        }

        outputPumper.start();
        errorPumper.start();

        final long startTime = System.currentTimeMillis();
        try {
            final boolean noTimeout = timeout <= 0;
            final long finish = noTimeout ? Long.MAX_VALUE : startTime + timeout;
            while (CommandLineUtils.isAlive(p)) {
                final long remainingTime = noTimeout ? Long.MAX_VALUE : finish - System.currentTimeMillis();
                if (remainingTime <= 0) break;
                final long maxWaitTime = quantum < remainingTime ? quantum : remainingTime;
                waitForProcess(maxWaitTime);
            }
            if (CommandLineUtils.isAlive(p)) {
                throw new InterruptedException("Process timeout out after " + timeout + " millis");
            }

            if (inputFeeder != null) {
                synchronized (inputFeeder) {
                    while (!inputFeeder.isDone()) {
                        inputFeeder.wait();
                    }
                }
            }

            synchronized (outputPumper) {
                while (!outputPumper.isDone()) {
                    outputPumper.wait();
                }
            }

            synchronized (errorPumper) {
                while (!errorPumper.isDone()) {
                    errorPumper.wait();
                }
            }

            return p.exitValue();
        } catch (InterruptedException ex) {
            p.destroy(); //was CommandLineUtils.killProcess(commandLine.getPid());
            throw new CommandLineTimeOutException(String.format("Error while executing external command, process killed after %d millis.", System.currentTimeMillis() - startTime), ex);
        }
        finally {
            if (inputFeeder != null) {
                inputFeeder.close();
            }
            outputPumper.close();
            errorPumper.close();
            p.destroy();
        }
    }

    /**
     * Extension point allowing subclasses to implement their own waiting.
     * Implementations should respect given maxWaitTime as the maximum time for which they can block execution.
     *
     * @param maxWaitTime maximum blocking time in milliseconds
     * @throws InterruptedException -
     */
    protected void waitForProcess(long maxWaitTime) throws InterruptedException {
        Thread.sleep(maxWaitTime);
    }

    /**
     * Sets the time limit for execution. Value less or equal to 0 cause no time limit.
     *
     * @param timeout
     */
    public void setTimeout(long timeout) {
        this.timeout = timeout;
    }

    public void setSystemIn(InputStream systemIn) {
        this.systemIn = systemIn;
    }

    public void setSystemOut(StreamConsumer systemOut) {
        this.systemOut = systemOut;
    }

    public void setSystemErr(StreamConsumer systemErr) {
        this.systemErr = systemErr;
    }

    public void setQuantum(long quantum) {
        this.quantum = quantum;
    }

    @Override
    public String toString() {
        return getClass().getSimpleName() + '{' + commandLine + '}';
    }
}
