package net.sf.buildbox.util.exec;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import org.codehaus.plexus.util.cli.Commandline;
import org.codehaus.plexus.util.cli.StreamConsumer;
import org.codehaus.plexus.util.cli.CommandLineException;

/**
 * Ensures that both stdout and stderr are processed in the same thread as caller.
 * Uses {@link BlockingQueue} internally to pass output lines to the main thread.
 */
public class SimpleExec extends CooperativeExec {
    private final BlockingQueue<DeviceLinePair> queue = new LinkedBlockingQueue<DeviceLinePair>();

    public SimpleExec(Commandline commandLine) {
        super(commandLine);
        super.setSystemOut(new MyStreamConsumer(true));
        super.setSystemErr(new MyStreamConsumer(false));
    }

    @Override
    public final void setSystemOut(StreamConsumer systemOut) {
        throw new IllegalAccessError();
    }

    @Override
    public final void setSystemErr(StreamConsumer systemOut) {
        throw new IllegalAccessError();
    }

    /**
     * Reads the process output queue and dispatches it to {@link #stdout(String)} and {@link #stderr(String)}.
     * It is not recommended to override this; in case of overriding, the descendant should call this ancestor implementation. 
     * @param maxWaitTime maximum blocking time in milliseconds
     * @throws InterruptedException
     */
    @Override
    protected void waitForProcess(long maxWaitTime) throws InterruptedException {
        final DeviceLinePair result = queue.poll(maxWaitTime, TimeUnit.MILLISECONDS);
        dispatch(result);
    }

    private void dispatch(DeviceLinePair result) throws InterruptedException {
        if (result == null) {
            //
        } else if (result.isStdOut) {
            stdout(result.line);
        } else {
            stderr(result.line);
        }
    }

    @Override
    public Integer call() throws CommandLineException {
        try {
            return super.call();
        } finally {
            // flush the queue
            while (!queue.isEmpty()) {
                try {
                    dispatch(queue.poll());
                } catch (InterruptedException e) {
                    System.err.println("interrupted while flushing: " + this);
                    Thread.currentThread().interrupt();
                }
            }
        }
    }

    protected void stdout(String line) throws InterruptedException {
        System.out.println(line);
    }

    protected void stderr(String line) throws InterruptedException {
        System.err.println(line);
    }

    private class MyStreamConsumer implements StreamConsumer {
        private final boolean isStdOut;

        public MyStreamConsumer(boolean stdOut) {
            isStdOut = stdOut;
        }

        public void consumeLine(String line) {
            queue.offer(new DeviceLinePair(isStdOut, line));
        }
    }

    private static class DeviceLinePair {
        boolean isStdOut;
        String line;

        private DeviceLinePair(boolean stdOut, String line) {
            isStdOut = stdOut;
            this.line = line;
        }
    }
}
