package net.sf.buildbox.util.props;

import java.lang.reflect.*;
import java.util.Properties;
import java.util.Map;
import java.util.HashMap;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import net.sf.buildbox.util.BbxStringUtils;

public class BbxProperties {
    private final Map<String, Object> values = new HashMap<String, Object>();
    private final Method[] methods;
    private final String prefix;

    private BbxProperties(final Method[] methods, final String prefix) {
        this.prefix = prefix;
        this.methods = methods;
    }

    private void readValues(Properties properties) {
        for (final Method method : methods) {
            final BbxProperty a = method.getAnnotation(BbxProperty.class);
            final String localName = a.name().equals("") ? method.getName() : a.name();
            final String key = this.prefix + localName;
            final String value;
            if (a.required()) {
                if (! properties.containsKey(key)) {
                    System.out.println("properties.keySet() = " + properties.keySet());
                    throw new IllegalArgumentException("Required property not set: " + key);
                }
                value = properties.getProperty(key);
            } else {
                value = properties.getProperty(key, a.defaultValue());
            }
            final Object result = value.equals("") ? null : convertToResultType(value, method);
            values.put(method.getName(), result);
        }
    }

    private Object get(String name) {
        if (! values.containsKey(name)) {
            throw new IllegalArgumentException("Method does not denote a property: " + name);
        }
        return values.get(name);
    }

    public static <T extends BbxPropertyConfig> T read(final Class<T> clazz, File propertyFile) throws IOException {
        final InputStream is = new FileInputStream(propertyFile.getAbsoluteFile());
        try {
            final Properties properties = new Properties();
            properties.load(is);
            return read(clazz, properties);
        } finally {
            is.close();
        }
    }

    /**
     * Reads given configuration, declared by clazz, from properties.
     * @param clazz interface class, public methods annottated with {@link net.sf.buildbox.util.props.BbxProperty}
     * @param properties the properties to take values from
     * @return instance allowing typesafe access to properties
     */
    @SuppressWarnings("unchecked")
    public static <T extends BbxPropertyConfig> T read(final Class<T> clazz, final Properties properties) {
        if (!clazz.isInterface()) {
            throw new IllegalArgumentException("not an interface: " + clazz);
        }
        // when BbxProperty is used on class, its name() attribute defines prefix. Otherwise, prefix is taken from package name
        final BbxProperty a = clazz.getAnnotation(BbxProperty.class);
        final String prefix = a == null ? clazz.getPackage().getName() + "." : a.name();
        final BbxProperties bbxProperties = new BbxProperties(clazz.getMethods(), prefix);
        final InvocationHandler invocationHandler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                // delegate Object method to proxy
                final Class<?> declaringClass = method.getDeclaringClass();
                if (declaringClass.equals(Object.class)) {
                    return method.invoke(proxy, args);
                }
                // fail on methods of any other ancestor
                if (!declaringClass.equals(clazz)) {
                    throw new IllegalStateException("cannot call method of other ancestors: " + method);
                }
                //
                return bbxProperties.get(method.getName());
            }

        };
        bbxProperties.readValues(properties);
        return (T) Proxy.newProxyInstance(clazz.getClassLoader(), new Class<?>[]{clazz}, invocationHandler);
    }

    private Object convertToResultType(String stringValue, Method method) {
        final Class<?> returnType = method.getReturnType();
        // check for some explicit types
        if (returnType.equals(String.class)) {
            return stringValue;
        }
        if (returnType.equals(Class.class)) {
            try {
                return Class.forName(stringValue);
            } catch (ClassNotFoundException e) {
                throw new IllegalArgumentException(e);
            }
        }
        if (returnType.equals(File.class)) {
            return new File(BbxStringUtils.expandSysProps(stringValue)); // TODO: for File only ?
        }
        // use static valueOf method if available (handles enums and basic types like Integer, Long, etc)
        try {
            try {
                final Constructor con = returnType.getConstructor(String.class);
                return con.newInstance(stringValue);
            } catch (NoSuchMethodException e) {
                //
            }
            final Method valueOfMethod = returnType.getMethod("valueOf", String.class);
            if (Modifier.isStatic(valueOfMethod.getModifiers()) && returnType.equals(valueOfMethod.getReturnType())) {
                return valueOfMethod.invoke(null, stringValue);
            }
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(e);
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException(e);
        } catch (InvocationTargetException e) {
            final Throwable ee = e.getCause();
            if (ee instanceof IllegalArgumentException) {
                throw (IllegalArgumentException) ee;
            } else {
                throw new IllegalArgumentException(ee);
            }
        } catch (InstantiationException e) {
            throw new IllegalStateException(e);
        }
        throw new IllegalStateException(String.format("unsupported return type '%s' on name method '%s'", returnType, method.getName()));
    }

}
